package com.atrium.jbpm;


import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import javax.annotation.PreDestroy;
import javax.naming.NamingException;
import javax.persistence.EntityManagerFactory;
import javax.persistence.Persistence;
import javax.transaction.TransactionManager;

import org.apache.log4j.Logger;
import org.drools.KnowledgeBase;
import org.drools.KnowledgeBaseFactory;
import org.drools.agent.KnowledgeAgent;
import org.drools.agent.KnowledgeAgentConfiguration;
import org.drools.agent.KnowledgeAgentFactory;
import org.drools.event.process.ProcessCompletedEvent;
import org.drools.event.process.ProcessEventListener;
import org.drools.event.process.ProcessNodeLeftEvent;
import org.drools.event.process.ProcessNodeTriggeredEvent;
import org.drools.event.process.ProcessStartedEvent;
import org.drools.event.process.ProcessVariableChangedEvent;
import org.drools.io.ResourceChangeScannerConfiguration;
import org.drools.io.ResourceFactory;
import org.drools.logger.KnowledgeRuntimeLoggerFactory;
import org.drools.persistence.jpa.JPAKnowledgeService;
import org.drools.runtime.Environment;
import org.drools.runtime.EnvironmentName;
import org.drools.runtime.KnowledgeSessionConfiguration;
import org.drools.runtime.StatefulKnowledgeSession;
import org.drools.runtime.process.ProcessInstance;
import org.jboss.seam.ScopeType;
import org.jboss.seam.annotations.Name;
import org.jboss.seam.annotations.Scope;
import org.jboss.seam.annotations.Startup;
import org.jbpm.process.audit.ProcessInstanceDbLog;
import org.jbpm.process.audit.ProcessInstanceLog;
import org.jbpm.process.audit.WorkingMemoryDbLogger;
import org.jbpm.workflow.instance.impl.WorkflowProcessInstanceImpl;


/**
 * Manage all the operations related to a process
 * @author mhaimoune
 *
 */
@Name("processManager")
@Startup
@Scope(ScopeType.APPLICATION)

public class ProcessManager {


	private static final ProcessManager instance = new ProcessManager();
	
	Logger log = Logger.getLogger(ProcessManager.class);
	

	/** entity manager factory to needed by jbpm to create the ksession */
	EntityManagerFactory emf;

//	private ProcessManager() {
//	};

	/** Singleton design pattern */
	public static ProcessManager getInstance() {

		return instance;
	}

	/** only one session is create to run all the process */
	private static StatefulKnowledgeSession ksession;

	/**
	 * Start a new process
	 * 
	 * @param process
	 * @param params
	 * @throws ProcessException 
	 */
	public final ProcessInstance startProcess(String process,
			Map<String, Object> params) {
		WorkflowProcessInstanceImpl processInstance = null;
		try {
			StatefulKnowledgeSession ksession = initSession();
			
			// start a new process instance
			processInstance = (WorkflowProcessInstanceImpl) ksession.startProcess(process, params);
			ksession.insert(processInstance);
			
			closeSession(ksession);

		} catch (Throwable t) {
			t.printStackTrace();
		}
		return processInstance;
	}

	
	/**
	 * signal an event to the process
	 * 
	 * @param type
	 * @param variable
	 * @param processInstanceId
	 * @throws ProcessException 
	 */
	public void signalEvent(String type, long processInstanceId) {
		initSession();
		ksession.signalEvent(type, null ,processInstanceId);
		closeSession(ksession);

	}


	/**
	 * Load the KnowledgeBase from the guvnor repository package - process
	 * definition - rules The package location can be changed in change-set.xml
	 */
	private KnowledgeBase readRepositoryKnowledgeBase() {

		/** Get the bpmn and rule files from repository after every build
		 * by starting a scanner witch check for notification each 5 min
		*/
		 ResourceChangeScannerConfiguration sconf = ResourceFactory.getResourceChangeScannerService().newResourceChangeScannerConfiguration(); 
		
		 sconf.setProperty(	 "drools.resource.scanner.interval", "90" );
		 ResourceFactory.getResourceChangeScannerService().configure( sconf );
		 ResourceFactory.getResourceChangeScannerService().start();
		 ResourceFactory.getResourceChangeNotifierService().start();
		 
		 
		 

		 // load the repository url from the change-set file
			KnowledgeAgentConfiguration aconf = KnowledgeAgentFactory
				.newKnowledgeAgentConfiguration();
		aconf.setProperty("drools.agent.newInstance", "false");
		KnowledgeAgent kagent = KnowledgeAgentFactory.newKnowledgeAgent("Guvnor default", aconf);
		kagent.applyChangeSet(ResourceFactory.newClassPathResource("change-set.xml"));
		KnowledgeBase kbase = kagent.getKnowledgeBase();
		 /*	
		KnowledgeBuilder kbuilder = KnowledgeBuilderFactory.newKnowledgeBuilder();
		kbuilder.add(ResourceFactory.newClassPathResource("ticket_process_V5.bpmn"), ResourceType.BPMN2);
		kbuilder.add(ResourceFactory.newClassPathResource("supplier_and_customer_rules.drl"), ResourceType.DRL);
		kbuilder.add(ResourceFactory.newClassPathResource("tasks_rules.drl"), ResourceType.DRL);
		KnowledgeBase	kbase = kbuilder.newKnowledgeBase();*/

		return kbase;
	}
	
	
	public StatefulKnowledgeSession newKsession (){
		
		// load the KnowledgeBase from the guvnor repository
		KnowledgeBase kbase = readRepositoryKnowledgeBase();

		EntityManagerFactory emf = Persistence
				.createEntityManagerFactory("org.jbpm.persistence.jpa");
		Environment env = KnowledgeBaseFactory.newEnvironment();

		env.set(EnvironmentName.ENTITY_MANAGER_FACTORY, emf);

		Properties sessionconfigproperties = new Properties();
		sessionconfigproperties
				.put("drools.processInstanceManagerFactory",
						"org.jbpm.persistence.processinstance.JPAProcessInstanceManagerFactory");
		sessionconfigproperties
				.put("drools.processSignalManagerFactory",
						"org.jbpm.persistence.processinstance.JPASignalManagerFactory");
		KnowledgeSessionConfiguration config = KnowledgeBaseFactory
				.newKnowledgeSessionConfiguration(sessionconfigproperties);
		
		env = KnowledgeBaseFactory.newEnvironment();
		env.set(EnvironmentName.ENTITY_MANAGER_FACTORY, emf);

		System.out.println("creating a new  session data ...");

		config = KnowledgeBaseFactory
				.newKnowledgeSessionConfiguration(sessionconfigproperties);
		ksession = JPAKnowledgeService
				.newStatefulKnowledgeSession(kbase, config,
						env);
		
		return ksession;
		
	}

	/**
	 * Get the ksession, create one of not yet created
	 * 
	 * @return
	 * @throws ProcessException 
	 */
	public StatefulKnowledgeSession initSession()  {

		if (ksession == null) {
			ksession = newStatefulKnowledgeSession();
		}
		return ksession;
	}
	
	
	/**
	 * check if the ksession is on
	 * 
	 * @return
	 * @throws ProcessException 
	 */
	public boolean isSessionOn() {

		if (ksession == null) {
			return false;
		}
		return true;
	}
	

	private void closeSession(StatefulKnowledgeSession ksession) {
		/*
		 * ksession.dispose() ; ksession = null;
		 */
	}
	
	public void dispose() {
		if (ksession!=null){
			System.out.println(" ksession dispose  ");
		  ksession.dispose() ; 
		  ksession = null;
		}else{
			System.out.println(" calling ksession dispose but not currant session was found ");
		}
	}

	/**
	 * create a new StatefulKnowledgeSession or load the existing one if it is
	 * already created
	 * 
	 * @return
	 * @throws ProcessException 
	 */
	private StatefulKnowledgeSession newStatefulKnowledgeSession(){
		
		
		StatefulKnowledgeSession ksession = null;
		try {

				System.out.println(" newStatefulKnowledgeSession  ");

			// load the KnowledgeBase from the guvnor repository
			KnowledgeBase kbase = readRepositoryKnowledgeBase();

		

			EntityManagerFactory emf = Persistence.createEntityManagerFactory("org.jbpm.persistence.jpa");
			Environment env = KnowledgeBaseFactory.newEnvironment();

			env.set(EnvironmentName.ENTITY_MANAGER_FACTORY, emf);

			Properties sessionconfigproperties = new Properties();
			sessionconfigproperties
					.put("drools.processInstanceManagerFactory",
							"org.jbpm.persistence.processinstance.JPAProcessInstanceManagerFactory");
			sessionconfigproperties
					.put("drools.processSignalManagerFactory",
							"org.jbpm.persistence.processinstance.JPASignalManagerFactory");
			KnowledgeSessionConfiguration config = KnowledgeBaseFactory
					.newKnowledgeSessionConfiguration(sessionconfigproperties);

			try {
				System.out.println("Loading session data ...");

				ksession = JPAKnowledgeService.loadStatefulKnowledgeSession(1,
						kbase, config, env);

			} catch (RuntimeException e) {
				if (e instanceof IllegalStateException) {
					Throwable cause = ((IllegalStateException) e).getCause();
					if (cause instanceof InvocationTargetException) {
						cause = cause.getCause();
						if (cause != null
								&& "Could not find session data for id 1"
										.equals(cause.getMessage())) {

							try {
								javax.naming.Context context = new javax.naming.InitialContext();
								TransactionManager tm = (javax.transaction.TransactionManager) context
										.lookup("java:/TransactionManager");
								tm.rollback();
							} catch (NamingException exception) {
								System.out.println("Unable to retrieve TransactionManager via JNDI: java:/TransactionManager : "
										+ exception.getMessage());

							}

							env = KnowledgeBaseFactory.newEnvironment();
							env.set(EnvironmentName.ENTITY_MANAGER_FACTORY, emf);

							System.out.println("creating a new  session data ...");

							config = KnowledgeBaseFactory
									.newKnowledgeSessionConfiguration(sessionconfigproperties);
							ksession = JPAKnowledgeService
									.newStatefulKnowledgeSession(kbase, config,
											env);
							  // add automatically inject the process instance in the working memory
						      // add a agendaEventListner , automatically fireAllRules 
								/*	((StatefulKnowledgeSessionImpl)  ((KnowledgeCommandContext) ((CommandBasedStatefulKnowledgeSession) ksession)
						                    .getCommandService().getContext()).getStatefulKnowledgesession() )
						                    .session.addEventListener(getEventListener() );*/

						} else {
							System.out.println("Error loading session data: " + cause);
							throw e;
						}
					} else {
						System.out.println("Error loading session data: " + cause);
						throw e;
					}
				} else {
					System.out.println("Error loading session data: " + e.getMessage());
					throw e;
				}
			}
			
            
			  ksession.addEventListener(getProcessEventListner());
			
			new WorkingMemoryDbLogger(ksession);
			 KnowledgeRuntimeLoggerFactory
				.newFileLogger(ksession, "process_log");
//			 KnowledgeRuntimeLoggerFactory
//				.newConsoleLogger(ksession);
			
			System.out.println("Successfully loaded default package from Guvnor");
			
		

		} catch (Throwable t) {
		t.printStackTrace();
		}
		
		return ksession;
	}

	/**
	 * this method add a variable to the current instance of a process
	 * 
	 * @param processInstanceId
	 *            process instance id
	 * @param varibleName
	 *            the variable name
	 * @param variableValues
	 *            the variable value
	 * @throws ProcessException 
	 */
	public void AddVariable(long processInstanceId, String variableName, Object variableValues) {
		
		System.out.println("Adding the varibale : " + variableName + ", Value "
				+ variableValues + " to the process instance "
				+ processInstanceId);
		
		StatefulKnowledgeSession ksession = initSession();

		WorkflowProcessInstanceImpl processInstance = (WorkflowProcessInstanceImpl) ksession
				.getProcessInstance(processInstanceId);

		if (processInstance!=null){
			
			processInstance.setVariable(variableName, variableValues);
		}else{
			System.out.println("Process instance : " + processInstanceId+ " Not active when inserting variable ");
		}	

		closeSession(ksession);
	}
	
	
	

	/**
	 * this method get a variable from the current instance
	 * 
	 * @param processInstanceId
	 *            process instance id
	 * @param varibleName
	 *            the variable name
	 * @param variableValue
	 *            the variable value
	 * @throws ProcessException 
	 */
	public Object getVariable(long processInstanceId, String variableName)  {
		StatefulKnowledgeSession ksession = initSession();
		Object vRetour = null;

			System.out.println(" getVariable : " + variableName );
		WorkflowProcessInstanceImpl processInstance = (WorkflowProcessInstanceImpl) ksession
				.getProcessInstance(processInstanceId);
		// if the instance existe and the process is active
		if (processInstance != null
				&& processInstance.getState() == ProcessInstance.STATE_ACTIVE) {
			vRetour = processInstance.getVariable(variableName);
			System.out.println("Varibale : " + variableName + " value : " + vRetour);
		} else {
			System.out.println(" process instance : " + processInstanceId+ " Not active when adding variable : " + variableName);
		}

		closeSession(ksession);
		return vRetour;
	}

	/**
	 * get the process instance using id
	 * @param processInstanceId
	 * @return
	 * @throws ProcessException 
	 */
	public ProcessInstance getProcessInstance(long processInstanceId)  {
		StatefulKnowledgeSession ksession = initSession();
		WorkflowProcessInstanceImpl processInstance = (WorkflowProcessInstanceImpl) ksession
				.getProcessInstance(processInstanceId);
		closeSession(ksession);
		return processInstance;

	}
	
	/**
	 * Get all the active process instance in the working memory of the ksession
	 * @throws ProcessException 
	 */
	public List <ProcessInstance> getActiveProcesses()  {
		StatefulKnowledgeSession ksession = initSession();
		List<ProcessInstanceLog> findActiveProcessInstances	= ProcessInstanceDbLog.findActiveProcessInstances("LoopTest");
		
		List<ProcessInstance> listProcess = new ArrayList<ProcessInstance>();
		if (findActiveProcessInstances != null ){
			System.out.println(" List active process log  :  " + findActiveProcessInstances.size());
			
			for (ProcessInstanceLog processInstanceLog : findActiveProcessInstances) {
				
				ProcessInstance p = ksession.getProcessInstance(processInstanceLog.getProcessInstanceId());
				if (p!=null){
					listProcess.add(p);
				}
				
			}
		}
		
		closeSession(ksession);
		
		return listProcess;

	}
	
	public void fireAllRules()  {
		StatefulKnowledgeSession ksession = initSession();
		ksession.fireAllRules();
		closeSession(ksession);

	}

	private  ProcessEventListener getProcessEventListner() {
		return new ProcessEventListener() {

			public void beforeProcessStarted(ProcessStartedEvent event) {
				// TODO Auto-generated method stub
				
			}

			public void afterProcessStarted(ProcessStartedEvent event) {
				// TODO Auto-generated method stub
				
			}

			public void beforeProcessCompleted(ProcessCompletedEvent event) {
				// TODO Auto-generated method stub
				
			}

			public void afterProcessCompleted(ProcessCompletedEvent event) {
				// TODO Auto-generated method stub
				
			}

			public void beforeNodeTriggered(ProcessNodeTriggeredEvent event) {
				// TODO Auto-generated method stub
				
			}

			public void afterNodeTriggered(ProcessNodeTriggeredEvent e) {
	            	//   System.out.println(" afterNodeTriggered : " + e.getNodeInstance().getNodeName() );
					
				// if rule task we fire all rules
				if ( e.getNodeInstance().getClass().getName().equals("org.jbpm.workflow.instance.node.RuleSetNodeInstance")){
					 System.out.println(" Fire all rules  for process instance: " + e.getProcessInstance().getId());
					// fire all the rules
					 try {
						fireAllRules();
					 }catch (Exception ex){

						 System.out.println("Exception   "  + ex.getMessage());
						 ex.printStackTrace();
					 
					 }
				}
				
			}

			public void beforeNodeLeft(ProcessNodeLeftEvent event) {
				// TODO Auto-generated method stub
				
			}

			public void afterNodeLeft(ProcessNodeLeftEvent event) {
				// TODO Auto-generated method stub
				
			}

			public void beforeVariableChanged(ProcessVariableChangedEvent event) {
				// TODO Auto-generated method stub
				
			}

			public void afterVariableChanged(ProcessVariableChangedEvent event) {
	            	//   System.out.println("afterVariableChanged : " + event );
				 org.drools.runtime.rule.FactHandle handle = event.getKnowledgeRuntime().getFactHandle( event.getProcessInstance() );
				 // we don't update when it's a tasks action, t to avoid infinite activation loop
              if(handle != null && !event.getVariableId().equals(JbpmConstantes.TASKS_ACTION_VARIABLE_NAME)){
             	 
             	 System.out.println("Injecting the process instance into the workin memory : " + event );
              
             	 event.getKnowledgeRuntime().update( handle, event.getProcessInstance() );
              
              }
			}
			
	
	};
	

	}
		
	/**
	 * If the application is restarted or stopped this method assure the ksession is disposed correctly
	 */
	@PreDestroy
	public void onDestroyProcessManager() {
		
		System.out.println(" onDestroy ProcessManager called to dipose the ksession  ");
		dispose();
		
	}

}
